#ifndef CONTACTO_INTER_RESIDUE_CONTACTS_FILTERING_H_
#define CONTACTO_INTER_RESIDUE_CONTACTS_FILTERING_H_

#include <vector>
#include <map>
#include <string>
#include <stdexcept>

#include "contact_id.h"

namespace contacto
{

template<typename ResidueIDType, typename ValueType>
std::map< contacto::ContactID<ResidueIDType>, ValueType > filter_inter_chain_contacts(const std::map< contacto::ContactID<ResidueIDType>, ValueType >& all_contacts)
{
	typedef std::map< contacto::ContactID<ResidueIDType>, ValueType > ContactsMap;
	ContactsMap interface_contacts;
	typename ContactsMap::iterator prev=interface_contacts.begin();
	for(typename ContactsMap::const_iterator it=all_contacts.begin();it!=all_contacts.end();++it)
	{
		if(it->first.a.chain_id!=it->first.b.chain_id)
		{
			prev=interface_contacts.insert(prev, *it);
		}
	}
	return interface_contacts;
}

template<typename ResidueIDType, typename ValueType, typename InterfaceReaderType>
std::map< contacto::ContactID<ResidueIDType>, ValueType > filter_inter_interval_contacts(const std::map< contacto::ContactID<ResidueIDType>, ValueType >& all_contacts, const std::string& intervals_string)
{
	typedef std::map< contacto::ContactID<ResidueIDType>, ValueType > ContactsMap;
	std::vector< std::vector< std::pair<ResidueIDType, ResidueIDType> > > intervals;
	if(!InterfaceReaderType::read_residue_ids_intervals(intervals_string, intervals) || intervals.size()<2)
	{
		throw std::runtime_error(std::string("Invalid intervals string: ")+intervals_string);
	}
	ContactsMap interface_contacts;
	typename ContactsMap::iterator prev=interface_contacts.begin();
	for(typename ContactsMap::const_iterator it=all_contacts.begin();it!=all_contacts.end();++it)
	{
		const ResidueIDType& a=it->first.a;
		const ResidueIDType& b=it->first.b;
		int a_group=-1;
		int b_group=-1;
		for(std::size_t i=0;i<intervals.size() && (a_group<0 || b_group<0);i++)
		{
			for(std::size_t j=0;j<intervals[i].size() && (a_group<0 || b_group<0);j++)
			{
				const ResidueIDType& r1=intervals[i][j].first;
				const ResidueIDType& r2=intervals[i][j].second;
				if(a_group<0 && a.chain_id==r1.chain_id && a.residue_number>=r1.residue_number && a.residue_number<=r2.residue_number)
				{
					a_group=i;
				}
				else if(b_group<0 && b.chain_id==r1.chain_id && b.residue_number>=r1.residue_number && b.residue_number<=r2.residue_number)
				{
					b_group=i;
				}
			}
		}
		if(a_group>=0 && b_group>=0 && a_group!=b_group)
		{
			prev=interface_contacts.insert(prev, *it);
		}
	}
	return interface_contacts;
}

template<typename ResidueIDType, typename ValueType, typename InterfaceReaderType>
void filter_custom_contacts(std::map< contacto::ContactID<ResidueIDType>, ValueType >& all_contacts, const bool inter_chain, const std::string& intervals_string)
{
	if(inter_chain)
	{
		all_contacts=filter_inter_chain_contacts<ResidueIDType, ValueType>(all_contacts);
	}

	if(!intervals_string.empty())
	{
		all_contacts=filter_inter_interval_contacts<ResidueIDType, ValueType, InterfaceReaderType>(all_contacts, intervals_string);
	}
}

}


#endif /* CONTACTO_INTER_RESIDUE_CONTACTS_FILTERING_H_ */
